/* asyndap.c - asynchronous search routines */

#ifndef lint
static char *rcsid = "$Header: /u1/home/isode/src/IC/IC-1/de/de/RCS/asyndap.c,v 9.3 1993/03/19 22:53:12 isode Exp isode $";
#endif

/*
 * $Header: /u1/home/isode/src/IC/IC-1/de/de/RCS/asyndap.c,v 9.3 1993/03/19 22:53:12 isode Exp isode $
 *
 *
 * $Log: asyndap.c,v $
 * Revision 9.3  1993/03/19  22:53:12  isode
 * Add QOS and DOS-DE code
 *
 * Revision 9.2  1992/12/15  07:05:26  awy
 * Update and fixes
 *
 * Revision 8.1  1991/09/13  14:36:41  isode
 * PARADISE Upgrade release
 *
 * Revision 8.0  91/07/17  13:18:48  isode
 * Release 7.0
 *
 *
 */

/*
 *				  NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions of a license agreement.
 *    Consult the Preface in the User's Manual for the full terms of
 *    this agreement.
 *
 */

#define asyndap_c
#ifdef DOS
#include <stdlib.h>
#include "proto-de.h"
#endif DOS
#include "messages.h"
#include <stdio.h>
#include <setjmp.h>
#ifdef USELDAP
#ifdef IC
#include <ldap/lber.h>
#include <ldap/ldap.h>
#else
#include <lber.h>
#include <ldap.h>
#endif /* IC */
#include <string.h>
#include "isomanifest.h"
#ifdef DOS
#include <msdos.h>
#else
#include <sys/time.h>
#endif DOS
#else
#include "quipu/dap2.h"
#include "quipu/ds_search.h"
#include "quipu/list.h"
#include "quipu/read.h"
#endif
#include "demanifest.h"
#include "asyn.h"
#include "util.h"

int opno = 0;
int opfloor; /* this is used as a floor beneath which any incoming results
                should be discarded - it is used to help with 
                abandoning operations */
static int opoutstanding = 0;

#ifdef USELDAP
extern LDAP * ldap_ld;
extern char * ldap_base;
extern int ldap_scope;
extern char ldap_filter[BUFSIZ];
extern char ** ldap_attrs;
extern char ldap_attrsonly;
extern int ldap_msgid;
extern int ldap_all;
extern struct timeval ldap_timeout;
extern LDAPMessage * ldap_res;
int nullint;
#endif USELDAP
extern jmp_buf abanbuf;
extern int dsap_ad;
int noOpsAbandoned;
struct namelist * abanlp = NULLLIST;
struct opsPending * ops = NULLOPL, * endops = NULLOPL;

int
countConns()
{
int count;
struct opsPending * oplp;

  for (oplp = ops , count = 0; oplp != NULLOPL; oplp = oplp->next, count++){};
  return count;
}

struct opsPending *
findOp(opnum)
int opnum;
{
struct opsPending * oplp;

  for (oplp = ops; oplp != NULLOPL; oplp = oplp->next)
  {
    if (oplp->opno == opnum)
      return oplp;
  }
  return NULLOPL;
}

#ifdef DOS
void
#else
int
#endif DOS
addToOpList(opnum, dnstr, type, parent)
int opnum;
char * dnstr;
int type;
struct opsPending * parent;
{
struct opsPending * oplp;

  oplp = oplist_alloc();
  oplp->opno = opnum;
  oplp->orgname = copy_string(dnstr);
  oplp->type = type; /* the sort of thing we are looking for */
  oplp->subordSearches = 0;
#ifdef QOS
  startQosTime(dnstr,&(oplp->qosRec));
#endif QOS
  oplp->parent = parent;
  oplp->lp = NULLLIST;
  oplp->next = NULLOPL;
  if (endops != NULLOPL)
    endops->next = oplp;
  else
    ops = oplp;
  endops = oplp;
}

/* routine returns the operation no. of the parent operation (e.g. the 
   search for OUs within an org), when all the subordinate searches (e.g. 
   the searches for people under those OUs) are complete.  This is to allow
   the parent search to be removed from the list, and the number of orgs
   still to be searched to be adjusted. */
int
removeFromOpList(opnum)
int opnum;
{
struct opsPending * oplp, * prev;
int retval;

  retval = 0;
  if (ops == NULLOPL)
  {
    (void)fprintf(stderr, REMOVEERR);
    return retval;
  }
  for (oplp = ops, prev = NULLOPL; oplp != NULLOPL; oplp = oplp->next)
  {
    if (oplp->opno != opnum)
      prev = oplp;
    else
      break;
  }
  if (oplp == NULLOPL)
  {
     (void)fprintf(stderr, NOOPERR, opnum);
     return retval;
  }
  /* if removing a person search operation from the list which has an ou
     parent, reduce searches outstanding under parent org */
  if (oplp->parent != NULLOPL)
    if (oplp->parent->subordSearches > 0)
    {
      if (--(oplp->parent->subordSearches) == 0)
        retval = oplp->parent->opno;
    }
    else
      (void)fprintf(stderr, "Problem with subord searches wanting to go negative\n");
  if (oplp == endops)
    endops = prev;
  if (prev != NULLOPL)
    prev->next = oplp->next;
  else
    ops = oplp->next;
  free(oplp->orgname);
  freeList(&(oplp->lp));
  free((char *)oplp);
  return retval;
}

#ifdef USELDAP
int
de_asynSearch(opnumber)
int * opnumber;
{
  int retval;

#ifdef DE-DEBUG
  print_search_args("de_asynSearch", ldap_base, ldap_scope, ldap_filter);
#endif DE-DEBUG
  retval = ldap_search(ldap_ld, ldap_base, ldap_scope, ldap_filter,
				ldap_attrs, ldap_attrsonly);
  if (retval == -1)
  {
    ldap_perror(ldap_ld, "de_asynSearch");
    return NOTOK;
  }
  *opnumber = retval;
  return OK;
}
#else
int
de_asynSearch(arg, error, opnumber)
struct ds_search_arg * arg;
struct DSError * error;
int * opnumber;
{
struct DAPindication di;

  di.di_type = 0;
  (void) DapSearch(dsap_ad, ++opno, arg, & di, ROS_ASYNC);
  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      *opnumber = -1;
      return NOTOK;
    case DI_ERROR:
      * opnumber = -1;
      *error = di.di_error.de_err; /* structure copy */
      return DS_ERROR_REMOTE;
    default:
      break;
  }
  *opnumber = opno;
  return OK;
}
#endif USELDAP

#ifdef USELDAP
int
de_search()
{
  int retval;

#ifdef DE-DEBUG
  print_search_args("de_search", ldap_base, ldap_scope, ldap_filter);
#endif DE-DEBUG
  retval = ldap_search(ldap_ld, ldap_base, ldap_scope, ldap_filter,
				ldap_attrs, ldap_attrsonly);
  if (retval == -1)
  {
    ldap_perror(ldap_ld, "de_search");
    return NOTOK;
  }
  ldap_msgid = retval;
  return de_wait_response(ldap_msgid, &nullint);
}
#else
int
de_search(arg, error, result)
struct ds_search_arg * arg;
struct DSError * error;
struct ds_search_result * result;
{

int ret;
struct DAPindication di;
struct DAPresult dapres;

  di.di_type = 0;
  opoutstanding = ++opno;
  ret = DapSearch(dsap_ad, opno, arg, & di, ROS_ASYNC);
  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      return NOTOK;
    case DI_ERROR:
      *error = di.di_error.de_err; /* structure copy */
      return DS_ERROR_REMOTE;
    default:
      break;
  }

  ret = de_wait_response(error, & dapres, opno, NULLIP);
  if (ret == OK)
    * result = dapres.dr_res.res_sr;
  return ret;
}
#endif USELDAP

#ifdef USELDAP
int
de_list()
{
  int retval;

#ifdef DE-DEBUG
  print_search_args("de_list", ldap_base, ldap_scope, ldap_filter);
#endif DE-DEBUG
  retval = ldap_search(ldap_ld, ldap_base, ldap_scope, ldap_filter,
				ldap_attrs, ldap_attrsonly);
  if (retval == -1)
  {
    ldap_perror(ldap_ld, "de_list");
    return NOTOK;
  }
  ldap_msgid = retval;
  return de_wait_response(ldap_msgid, &nullint);
}
#else
int
de_list(arg, error, result)
struct ds_list_arg * arg;
struct DSError * error;
struct ds_list_result * result;
{

int ret;
struct DAPindication di;
struct DAPresult dapres;

  di.di_type = 0;
  opoutstanding = ++opno;
  ret = DapList(dsap_ad, opno, arg, & di, ROS_ASYNC);
  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      return NOTOK;
    case DI_ERROR:
      *error = di.di_error.de_err; /* structure copy */
      return DS_ERROR_REMOTE;
    default:
      break;
  }

  ret = de_wait_response(error, & dapres, opno, NULLIP);
  if (ret == OK)
    * result = dapres.dr_res.res_ls;
  return ret;
}
#endif USELDAP

#ifdef USELDAP
int
de_read()
{
  int retval;

#ifdef DE-DEBUG
  print_search_args("de_read", ldap_base, ldap_scope, ldap_filter);
#endif DE-DEBUG
  retval = ldap_search(ldap_ld, ldap_base, ldap_scope, ldap_filter,
				ldap_attrs, ldap_attrsonly);
  if (retval == -1)
  {
    ldap_perror(ldap_ld, "de_read");
    return NOTOK;
  }
  ldap_msgid = retval;
  return de_wait_response(ldap_msgid, &nullint);
}
#else
int
#if (OSISEC2 && IC)
de_read(arg, error, result, signedop)
#else
de_read(arg, error, result)
#endif
struct ds_read_arg * arg;
struct DSError * error;
struct ds_read_result * result;
#if (OSISEC2 && IC)
int signedop;
#endif
{

int ret;
struct DAPindication di;
struct DAPresult dapres;

  di.di_type = 0;
  opoutstanding = ++opno;
#if (OSISEC2 && IC)
  ret = DapRead_secure(dsap_ad, opno, arg, & di, ROS_ASYNC, signedop);
#else
  ret = DapRead(dsap_ad, ++opno, arg, & di, ROS_ASYNC);
#endif
  switch(di.di_type)
  {
    case DI_PREJECT:
    case DI_ABORT:
      return NOTOK;
    case DI_ERROR:
      *error = di.di_error.de_err; /* structure copy */
      return DS_ERROR_REMOTE;
    default:
      break;
  }
  ret = de_wait_response(error, & dapres, opno, NULLIP);
  if (ret == OK)
    * result = dapres.dr_res.res_rd;
#ifdef OSISEC
  if (signedop == DE_SIGNED)
  {
    if (!result->rdr_common.cr_security)
    {
      resetprint(NOPROTRESULTS);
      return NOTOK;
    }
    else if (!result->rdr_common.cr_security->sp_name)
    {
      resetprint(DUANOTIDENTIFIED);
      return NOTOK;
    }
    else if (dn_cmp(arg->rda_common.ca_requestor, 
		result->rdr_common.cr_security->sp_name) != 0)
    {
      resetprint(WRONGDUA);
      return NOTOK;
    }
    else if (dn_cmp(result->rdr_common.cr_requestor, 
		arg->rda_common.ca_security->sp_name) != 0)
    {
      resetprint(WRONGDSA);
      return NOTOK;
    }
  }
#endif OSISEC
  return ret;
}
#endif USELDAP

#ifdef USELDAP
int
waitSearchResponse(opnum)
int * opnum;
{
  return de_wait_response(LDAP_RES_ANY, opnum);
}
#else
int
waitSearchResponse(err, res, opnum)
struct DSError * err;
struct ds_search_result * res;
int * opnum;
{
struct DAPresult dapres;
int ret;

  ret = de_wait_response(err, & dapres, 0, opnum); /* 0 means any op will do */
  if (ret == OK)
  {
    * res = dapres.dr_res.res_sr;
  }
  return ret;
}
#endif USELDAP

#ifdef USELDAP
int
de_wait_response(opwant, opgot)
int opwant;
int * opgot;
{
  int res;

/*  if (ldap_res != (LDAPMessage *)NULL)
    ldap_msgfree(ldap_res);*/
  while ((res = ldap_result(ldap_ld, opwant, ldap_all, &ldap_timeout,
		&ldap_res)) == 0)
#ifdef DOS
  {
    checkCtrlC();
    doInProgress();
  }
#else
    doInProgress();
#endif DOS
  ldap_msgid = 0;
  if (res < 0)
  {
    ldap_perror(ldap_ld, "de_wait_response");
    return NOTOK;
  }

  switch(ldap_result2error(ldap_ld, ldap_res, 0))
  {
  case LDAP_SUCCESS:
  case LDAP_TIMELIMIT_EXCEEDED:
  case LDAP_SIZELIMIT_EXCEEDED:
    /* We'll find out whether the TIMELIMIT or SIZELIMIT have been exceeded
       later by looking at ldap_ld->errno. */
    *opgot = ldap_res->lm_msgid;
    return OK;
  default:
    *opgot = ldap_res->lm_msgid;
    return NOTOK;
  }
  /* NOTREACHED */
}
#else
int
de_wait_response(err, dapres, opwant, opgot)
struct DSError * err;
struct DAPresult * dapres;
int opwant;
int * opgot;
{
struct DAPindication di;

  while (1)
  {
    (void) DapInitWaitRequest(dsap_ad, 2, &di);
    switch(di.di_type)
    {
      case DI_RESULT:
         if (((opwant != 0) && (di.di_result.dr_id != opwant)) ||
               (di.di_result.dr_id <= opfloor))
	 {
           /* purpose of this is to throw away any rubbish we don't want */
	   continue;
	 }
	 opoutstanding = 0;
         *dapres = di.di_result; /* structure copy */
	 if (opgot)
	   *opgot = di.di_result.dr_id;
         return OK;
      case DI_ERROR:
         if (((opwant != 0) && (di.di_error.de_id != opwant)) ||
               (di.di_result.dr_id <= opfloor))
	 {
           /* purpose of this is to throw away any rubbish we don't want */
	   continue;
	 }
	opoutstanding = 0;
        *err = di.di_error.de_err; /* structure copy */
	if (opgot)
	  *opgot = di.di_error.de_id;
        return DS_ERROR_REMOTE;
      case DI_PREJECT:
	opoutstanding = 0;
        if (opgot)
          *opgot = di.di_preject.dp_id;
        return PROTREJECT;
      case DI_ABORT:
        doInProgress();
        continue;
    }
  }
}
#endif USELDAP

void
initAbans()
{
  noOpsAbandoned = 0;
  freeList(&abanlp);
}

void
#ifdef DOS
abandonops(sig)
int sig;
#else
abandonops()
#endif DOS
{
#ifndef USELDAP
struct ds_abandon_arg arg;
struct DAPindication di;
#endif USELDAP
extern int goneAsync;
struct opsPending * oplp, * next;

  resetprint(ABANSEARCHES);
  noOpsAbandoned = 0;
  for (oplp = ops; oplp != NULLOPL;)
  {
#ifdef USELDAP
      ldap_abandon(ldap_ld, oplp->opno);
#else
      arg.aba_invokeid = oplp->opno;
      (void) DapAbandon(dsap_ad, ++opno, &arg, &di, ROS_SYNC);
#endif USELDAP
#ifdef QOS
      drawstar();
      updateStats(QOS_ABANDON, &(oplp->qosRec));
#endif QOS
      addToList(&abanlp, orgNameFromDN(oplp->orgname, 1));
      noOpsAbandoned++;
      free(oplp->orgname);
      next = oplp->next;
      free((char *)oplp);
      oplp = next;
  }
#ifdef QOS
  (void)printf(" \b");
#endif QOS
  (void)printf("\n");
  goneAsync = FALSE;
  ops = endops = NULLOPL;
  opfloor = opno;
  longjmp(abanbuf, 0);
}

#ifndef USELDAP
de_abandon()
{
struct ds_abandon_arg arg;
struct DAPindication di;

/* maybe should try and see if an operation has been invoked or not,
   before trying to abandon it */

   if (opoutstanding != 0)
   {
     arg.aba_invokeid = opno;;
     (void) DapAbandon(dsap_ad, ++opno, &arg, &di, ROS_SYNC);
   }
}
#endif

#ifdef DOS
void
#endif DOS
printOrgAbandons()
{
struct namelist * lp;

  pageprint(ABANORGS);
  for (lp = abanlp; lp != NULLLIST; lp = lp->next)
      pageprint("  %s\n", lp->name);
}


#ifdef USELDAP
#ifdef DE-DEBUG
print_search_args(caller, base, scope, filt)
char * caller, * base, * filt;
{
char str[BUFSIZ];

  switch (scope) {
  case LDAP_SCOPE_BASE: (void)strcpy(str, "base");
			break;
  case LDAP_SCOPE_ONELEVEL: (void)strcpy(str, "onelevel");
			break;
  case LDAP_SCOPE_SUBTREE: (void)strcpy(str, "subtree");
			break;
  }
  (void)fprintf(stderr, "%s : %s : %s : %s\n", caller, base, str, filt);
}
#endif DE-DEBUG
#endif USELDAP
