/* loc.c - search for a locality */

#ifndef lint
static char *rcsid = "$Header: /u1/home/isode/src/IC/IC-1/de/de/RCS/loc.c,v 9.3 1993/03/19 22:54:50 isode Exp isode $";
#endif

/*
 * $Header: /u1/home/isode/src/IC/IC-1/de/de/RCS/loc.c,v 9.3 1993/03/19 22:54:50 isode Exp isode $
 *
 *
 * $Log: loc.c,v $
 * Revision 9.3  1993/03/19  22:54:50  isode
 * Add QOS and DOS-DE code
 *
 * Revision 9.2  1992/12/15  07:05:26  awy
 * Update and fixes
 *
 * Revision 9.1  1992/11/12  20:05:23  awy
 * IC-1 base
 *
 */

/*
 *				  NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions of a license agreement.
 *    Consult the Preface in the User's Manual for the full terms of
 *    this agreement.
 *
 */

#define loc_c
#ifdef DOS
#include <stdlib.h>
#include "proto-de.h"
#endif DOS
#include "messages.h"
#include <signal.h>
#ifdef USELDAP
#include <string.h>
#ifdef IC
#include <ldap/lber.h>
#include <ldap/ldap.h>
#else
#include <lber.h>
#include <ldap.h>
#endif /* IC */
#include "isomanifest.h"
#ifdef DOS
#include <msdos.h>
#else
#include <sys/time.h>
#endif DOS
#else
#include "quipu/util.h"
#include "quipu/read.h"
#include "quipu/list.h"
#include "quipu/ds_search.h"
#include "quipu/dua.h"
#include "util.h"
#endif USELDAP
#include "demanifest.h"
#include "namelist.h"
#include "filter.h"
#include "destrings.h"
#include "query.h"

#ifdef USELDAP
extern LDAP * ldap_ld;
extern char * ldap_base;
extern int ldap_scope;
extern char ldap_filter[BUFSIZ];
extern char ** ldap_attrs;
extern char ldap_attrsonly;
extern int ldap_msgid;
extern int ldap_all;
extern struct timeval ldap_timeout;
extern LDAPMessage * ldap_res;
LDAPMessage * ldap_delete_result_entry();
#else
extern char filbuf[];
extern struct ds_list_arg larg;
extern struct ds_list_result lresult;

extern struct ds_search_arg sarg;
extern struct ds_search_result sresult;
extern struct DSError serror;
#endif USELDAP
extern struct query qinfo[];

extern int abandoned;
extern int highNumber;
extern int numberType;
extern int exactMatch;
extern int fuzzyMatching;
extern int gotFuzzyMatch;
extern int quipuMastersCo;
extern char exactString[];
extern int searchAliasLo;
extern char searchVal[];

extern struct namelist * atts[];

#ifdef USELDAP
void fillMostLocListArgs();
void fillMostLocSearchArgs();
#else
struct ds_list_arg *fillMostLocListArgs();
struct ds_search_arg *fillMostLocSearchArgs();
#endif USELDAP

void makeExplicitLocFilter();
void locFilter1(), locFilter2(), locFilter3(), locFilter4();


VFP explicitLoc[] = {genExplicitFilter, NULLVFP};
VFP normalLoc[] = {genFilter1, genFilter2, genFilter3, genFilter4, NULLVFP};

int
listLocs(cstr, lstr, llistp)
char * cstr, * lstr;
struct namelist ** llistp;
{
	 return (listEntries(cstr, lstr, llistp, LOCALITY));
}

void
printListLocs(loc, llistp)
char * loc;
struct namelist * llistp;
{
struct namelist * x;
int i;

	if (llistp == NULLLIST)
		if (strcmp(loc, "*") == 0)
			pageprint(NOLOCSFOUND);
		else
			pageprint(NOLOCSMATCH, loc);
	else
	{
		for (i = 1, x = llistp; x != NULLLIST; i++, x = x->next)
			printLastComponent(INDENTON, x->name, LOCALITY, i);
		showAnyProblems(loc);
	}
}

#ifdef BUM
int
listAllLocs(cstr, llistp)
char * cstr;
struct namelist ** llistp;
{
int ret;

	if (quipuMastersCo == FALSE)
	{
#ifdef USELDAP
		fillMostLocListArgs(cstr);
#else
		larg = * fillMostLocListArgs(cstr);
#endif USELDAP
		ret = reallyMakeListLocs(llistp);
	}
	else
	{
#ifdef USELDAP
	        fillMostLocSearchArgs(cstr, LDAP_SCOPE_ONELEVEL);
	        genAllFilter(DE_LOCALITY);
#else
	        sarg = * fillMostLocSearchArgs(cstr, SRA_ONELEVEL);
	        genAllFilter(DE_LOCALITY, &sarg.sra_filter);
#endif USELDAP
	        ret = makeListLocalities(llistp);
	}
        if (ret != OK)
		logListSuccess(LIST_ERROR, "loc", 0);
	else
		logListSuccess(LIST_OK, "loc", listlen(*llistp));
	if (quipuMastersCo == FALSE)
		freeListArgs();
	else
		freeSearchArgs();
	alarmCleanUp();
	return ret;
}

int
listMatchingLocs(cstr, lstr, llistp)
char * cstr, * lstr;
struct namelist ** llistp;
{
VFP * filtarray;
VFP filterfunc;
int filtnumber;
char rstr[BUFSIZ];

        if (index(lstr, '*') != NULLCP) /* contains at least one asterisk */
	{
                filtarray = explicitLoc;
		filtnumber = -1;
	}
        else
	{
                filtarray = normalLoc;
		filtnumber = 0;
	}
	if (quipuMastersCo ==FALSE) /* try read first */
	{
#ifdef USELDAP
		(void)sprintf(rstr, "l=%s, %s", lstr, cstr);
#else
		(void)sprintf(rstr, "%s@l=%s", cstr, lstr);
#endif USELDAP
		if (readExactLoc(rstr, llistp) == OK)
		{
			logReadSuccess(READ_OK, "loc");		
			return OK;
		}
		else
			/* else try searches */
			logReadSuccess(READ_FAIL, "loc");
	}
#ifdef USELDAP
	fillMostLocSearchArgs(cstr, LDAP_SCOPE_ONELEVEL);
#else
	sarg = * fillMostLocSearchArgs(cstr, SRA_ONELEVEL);
#endif USELDAP
        while ((filterfunc = *filtarray++) != NULLVFP)
	{
		filtnumber++;
#ifdef USELDAP
                filterfunc(lstr, DE_LOCALITY, DE_LOCALITY_NAME);
		if (strcmp(ldap_filter, "") == 0)
			continue;
#else
                filterfunc(lstr, DE_LOCALITY, DE_LOCALITY_NAME,
			&sarg.sra_filter);
		if (sarg.sra_filter == NULLFILTER)
			continue;
#endif USELDAP
                if (makeListLocalities(llistp) != OK)
		{
			freeSearchArgs();
			logSearchSuccess(SEARCH_ERROR, "loc", lstr, filtnumber, 0);
			alarmCleanUp();
			return NOTOK;
		}
                if (*llistp != NULLLIST)
                        break;
	}
	if (*llistp != NULLLIST)
		logSearchSuccess(SEARCH_OK, "loc", lstr, filtnumber, listlen(*llistp));
	else
		logSearchSuccess(SEARCH_FAIL, "loc", lstr, filtnumber, 0);
	freeSearchArgs();
	alarmCleanUp();
	return OK;
}

int
readExactLoc(lstr, llistp)
char * lstr;
struct namelist ** llistp;
{
int ret;

	ret = readLoc(lstr, llistp);
	alarmCleanUp();
	return ret;
}

int
reallyMakeListLocs(llistp)
struct namelist ** llistp;
{
#ifdef USELDAP
LDAPMessage * enp, *tmpenp;
#else
struct subordinate * x;
DN dn1, dn2;
char * cp;
#endif USELDAP
int retval;

	if (rebind() != OK)
		return NOTOK;
#ifdef USELDAP
	retval = de_list();
	/* something should happen here ??? */
#else
	retval = de_list(&larg, &serror, &lresult);
	if ((retval == DSE_INTR_ABANDONED) &&
	    (serror.dse_type == DSE_ABANDONED))
	       abandoned = TRUE;
#endif USELDAP

	/* setProblemFlags(sresult); */
	highNumber = 0;
#ifdef USELDAP
	if (ldap_count_entries(ldap_ld, ldap_res) > 0)
	    for (enp = ldap_first_entry(ldap_ld, ldap_res); enp != NULL;
				enp = tmpenp)
	    {
		tmpenp = ldap_next_entry(ldap_ld, enp);
		*llistp = list_alloc();
		(*llistp)->name = ldap_get_dn(ldap_ld, enp);
		(*llistp)->ats = ldap_delete_result_entry( &ldap_res, enp );
		llistp = &(*llistp)->next;
		highNumber++;
	    }
	freeLdapMsg(ldap_res);
#else
	for (x = lresult.lsr_subordinates; x != (struct subordinate *) NULL;
	                                    x = x->sub_next)
	{
		/* only add locs to the list */
		cp = strdup(attr2name(x->sub_rdn->rdn_at, OIDPART));
		if (strcmp(cp, DE_LOCALITY_NAME) == 0)
		{
			*llistp = list_alloc();
			dn1 = dn_cpy(larg.lsa_object);
			dn2 = dn_comp_new(x->sub_rdn);
			dn_append(dn1, dn2);
			(*llistp)->name = dn2pstr(dn1);
			(*llistp)->ats = NULLATTR;
			llistp = &(*llistp)->next;
			highNumber++;
			
		}
		free(cp);
	}
#endif USELDAP
	*llistp = NULLLIST;
#ifndef USELDAP
	/* free something */
	dn_free (lresult.lsr_object);
	/* free something else */
#endif USELDAP
	return OK;
}

int
makeListLocalities(llistp)
struct namelist ** llistp;
{
#ifdef USELDAP
LDAPMessage * enp, * tmpenp;
#else
entrystruct * x;
#endif USELDAP
int retval;

	if (rebind() != OK)
		return NOTOK;
#ifdef USELDAP
	retval = de_search();
#else
	retval = de_search(&sarg, &serror, &sresult);
	if ((retval == DSE_INTR_ABANDONED) &&
	    (serror.dse_type == DSE_ABANDONED))
		abandoned = TRUE;
#endif USELDAP
        if (retval != OK)
                return NOTOK;
#ifdef USELDAP
	setProblemFlags(ldap_ld->ld_errno);
#else
	correlate_search_results (&sresult);

	setProblemFlags(sresult);
#endif USELDAP

	highNumber = 0;
#ifdef USELDAP
	if (ldap_count_entries(ldap_ld, ldap_res) > 0)
	    for (enp = ldap_first_entry(ldap_ld, ldap_res); enp != NULL;
			enp = tmpenp)
	    {
		tmpenp = ldap_next_entry(ldap_ld, enp);
		*llistp = list_alloc();
		(*llistp)->name = ldap_get_dn(ldap_ld, enp);
		(*llistp)->ats = ldap_delete_result_entry( &ldap_res, enp );
		llistp = &(*llistp)->next;
		highNumber++;
	    }
	freeLdapMsg(ldap_res);
#else
	for (x = sresult.CSR_entries; x != NULLENTRYINFO; x = x->ent_next) 
	{
		*llistp = list_alloc();
		(*llistp)->name = dn2pstr(x->ent_dn);
                (*llistp)->ats = as_cpy(x->ent_attr);
		llistp = &(*llistp)->next;
		highNumber++;
	}
#endif USELDAP
	*llistp = NULLLIST;
#ifndef USELDAP
	entryinfo_free(sresult.CSR_entries, 0);
	dn_free (sresult.CSR_object);
	crefs_free (sresult.CSR_cr);
	filter_free(sarg.sra_filter);
#endif USELDAP
	return OK;
}

int
readLoc(lstr, llistp)
char * lstr;
struct namelist ** llistp;
{
#ifdef USELDAP
LDAPMessage * enp;
char * cp;
#else
static struct ds_read_arg rarg;
static struct ds_read_result rresult;
static struct DSError rerror;
static CommonArgs sca = default_common_args;
Attr_Sequence * atl;
AttributeType at;
struct namelist * x;
#endif USELDAP
int retval;

	if (rebind() != OK)
		return NOTOK;
#ifdef USELDAP
	ldap_base = removeLastRDN(lstr);
	ldap_scope = LDAP_SCOPE_ONELEVEL;
	ldap_attrs = makeAllAttrs(atts[LOCALITY]);
	ldap_attrsonly = 0;
	ldap_all = 1;
	ldap_timeout.tv_sec = 2;
	ldap_timeout.tv_usec = 0;
	(void)sprintf(ldap_filter, "(%s)", (cp = lastRDN(lstr)));
	free(cp);
	retval = de_read();
	freeLdapOpArgs();
#else
	rarg.rda_common =sca; /* struct copy */
	rarg.rda_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	rarg.rda_common.ca_servicecontrol.svc_sizelimit= SVC_NOSIZELIMIT;
	rarg.rda_object = str2dn(lstr);
	/* specify attributes of interest */
	rarg.rda_eis.eis_allattributes = FALSE;
	atl = &(rarg.rda_eis.eis_select);
	for (x = atts[LOCALITY]; x != NULLLIST; x = x->next)
	{
		if ((at = str2AttrT(x->name)) == NULLAttrT)
			continue;
		*atl = as_comp_alloc();
		(*atl)->attr_type = at;
		(*atl)->attr_value = NULLAV;
		atl = &(*atl)->attr_link;
	}
	*atl = NULLATTR;
	rarg.rda_eis.eis_infotypes = EIS_ATTRIBUTESANDVALUES;
	retval = de_read(&rarg, &rerror, &rresult, DE_UNSIGNED);
	if ((retval == DSE_INTR_ABANDONED) &&
            (rerror.dse_type == DSE_ABANDONED))
                abandoned = TRUE;
#endif USELDAP
	if (retval != OK)
	        return NOTOK;
	/* setProblemFlags(sresult); */
	highNumber = 1;	
	*llistp = list_alloc();
#ifdef USELDAP
	if (ldap_count_entries(ldap_ld, ldap_res) > 0) {
		enp = ldap_first_entry(ldap_ld, ldap_res);
		(*llistp)->name = ldap_get_dn(ldap_ld, enp);
		(*llistp)->ats = ldap_delete_result_entry( &ldap_res, enp );
	}
	else
		return NOTOK;
	freeLdapMsg(ldap_res);
#else
	(*llistp)->name = dn2pstr(rresult.rdr_entry.ent_dn);
	(*llistp)->ats = as_cpy(rresult.rdr_entry.ent_attr);
#endif USELDAP
	(*llistp)->next = NULLLIST;
	return retval;
}

#ifdef USELDAP
void
fillMostLocListArgs(str)
char * str;
{
	ldap_base = copy_string(str);
	ldap_scope = LDAP_SCOPE_ONELEVEL;
	ldap_attrs = (char **) NULL;
	ldap_attrsonly = 1;
	ldap_all = 1;
	ldap_timeout.tv_sec = 2;
	ldap_timeout.tv_usec = 0;
	(void)sprintf(ldap_filter, "(objectclass=locality)");
}
#else
struct ds_list_arg *
fillMostLocListArgs(str)
char * str;
{
static struct ds_list_arg arg;
static CommonArgs sca = default_common_args;

	arg.lsa_common =sca; /* struct copy */
	arg.lsa_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	arg.lsa_common.ca_servicecontrol.svc_sizelimit= SVC_NOSIZELIMIT;

	arg.lsa_object = str2dn(str);
	return (&arg);
}
#endif USELDAP

#ifdef USELDAP
void
fillMostLocSearchArgs(cstr, searchdepth)
char * cstr;
int searchdepth;
{
	if (strcmp(cstr, "root") == 0)
		ldap_base = copy_string("");
	else
		ldap_base = copy_string(cstr);
	ldap_scope = searchdepth;
	ldap_attrs = makeAllAttrs(atts[LOCALITY]);
	ldap_attrsonly = 0;
	ldap_all = 1;
	ldap_timeout.tv_sec = 2;
	ldap_timeout.tv_usec = 0;
}
#else
struct ds_search_arg *
fillMostLocSearchArgs(cstr, searchdepth)
char * cstr;
int searchdepth;
{
static struct ds_search_arg arg;
Attr_Sequence * atl;
AttributeType at;
struct namelist * x;
static CommonArgs sca = default_common_args;

	arg.sra_common = sca; /* struct copy */
        arg.sra_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
        arg.sra_common.ca_servicecontrol.svc_sizelimit= SVC_NOSIZELIMIT;

	arg.sra_subset = searchdepth;
	if (strcmp(cstr, "root") == 0)
		arg.sra_baseobject = NULLDN;
	else
		arg.sra_baseobject = str2dn(cstr);
	arg.sra_searchaliases = searchAliasLo;
	/* specify attributes of interest */
	arg.sra_eis.eis_allattributes = FALSE;
	atl = &(arg.sra_eis.eis_select);
	for (x = atts[LOCALITY]; x != NULLLIST; x = x->next)
	{
		if ((at = str2AttrT(x->name)) == NULLAttrT)
			continue;
		*atl = as_comp_alloc();
		(*atl)->attr_type = at;
		(*atl)->attr_value = NULLAV;
		atl = &(*atl)->attr_link;
	}
	*atl = NULLATTR;
	arg.sra_eis.eis_infotypes = EIS_ATTRIBUTESANDVALUES;
	return (&arg);
}
#endif USELDAP
#endif BUM
