#ifdef USELDAP
#ifdef IC
#include <ldap/lber.h>
#include <ldap/ldap.h>
#else
#include <lber.h>
#include <ldap.h>
#endif /* IC */
#include <string.h>
#endif

#include "util.h"
#include "demanifest.h"
#include "destrings.h"
#include "namelist.h"

#ifdef USELDAP
extern LDAP * ldap_ld;
#endif

struct sortel {
    struct namelist *lp;
    char * key;
};

static struct sortel * sortbufp;

static int
compareKeys(x, y)
struct sortel * x, * y;
{
	return strcmp(x->key, y->key);
}

char *
getprrname(lp)
struct namelist * lp;
{
#ifdef USELDAP
BerElement *cookie;
char * at;
char **av;
static char surname[BUFSIZ];
#else
Attr_Sequence at;
extern char * val2str();
#endif

#ifdef USELDAP
    for (at = ldap_first_attribute(ldap_ld, lp->ats, &cookie);
	at != NULL; at = ldap_next_attribute(ldap_ld, lp->ats, cookie))
    {
	if (strcmp(at, SHORT_SN) == 0)
	{
	    av = ldap_get_values(ldap_ld, lp->ats, at);
	    (void) strcpy(surname, av[0]);
	    ldap_value_free(av);
	    return surname;
	}
    }
#else
    for (at = lp->ats; at != NULLATTR; at = at->attr_link)
    {
	if (strcmp(attr2name(at->attr_type, OIDPART), DE_SURNAME) == 0)
	{
	    return val2str(&(at->attr_value->avseq_av));
	}
    }
/* and a simpler, less general version which was usweful during testing
	return (rindex(lp->name, ' ') + 1); */
#endif
    /* if there isn't a surname return the common name */
    return lastComponent(lp->name, PERSON);
}

/* dept is used to indicate whether the search was under a dept or not
   if not, need to extract a dept from the name in order to sort the results
   by dept */
char *
makeKey(lp, objectType, dept)
struct namelist * lp;
int objectType;
int dept;
{
char key[BUFSIZ];
char * cp, * cp2;

    switch(objectType)
    {
        case PERSON:
	    (void)strcpy(key, " \n"); /* default dept value to a space if no dept */
            if (dept == FALSE)
            {
                cp = removeLastRDN(lp->name);
#ifdef USELDAP
		if (index(cp, ',') != rindex(cp, ','))
		{
		    if (strncmp(cp, SHORT_OU, strlen(SHORT_OU)) == 0)
#else
		if (index(cp, '@') != rindex(cp, '@'))
		{
		    if (strncmp((rindex(cp, '@') + 1),
			SHORT_OU, strlen(SHORT_OU)) == 0)
#endif
		    {
			    (void)strcpy(key, (cp2 = lastComponent(cp, ORGUNIT)));
			    free(cp2);
			    (void)strcat(key, "\n"); /* newline is dept / surname 
						separator char */
		    }
		}
		free(cp);
	    }
	    (void)strcat(key, (cp2 = getprrname(lp)));
	    free(cp2);
	    break;
        default:
            /* this is no more than place filler code - the code
               is intended to be generalised, but it has only been
               tested for personal names */
	    (void)strcpy(key, lastComponent(lp->name, objectType));
    }
    return strdup(key);
}

void
reSeqList(lp, len)
struct namelist ** lp;
int len;
{
struct namelist * x;
int i;
struct sortel * sbp;

	sbp = sortbufp;
	*lp = x = sbp->lp;
	for (i = 1, sbp++; i < len; i++, sbp++)
	{
		x->next = sbp->lp;
		x = x->next;
	}
	x->next = NULLLIST;
}

void
sortList(objectType, lp, pstr)
int objectType;
struct namelist ** lp;
char * pstr;
{
int i, len;
struct namelist * x;
struct sortel * sbp;

int dept;
	
	len = listlen(*lp);
	if (len < 2)
		return;
	sortbufp = (struct sortel *)malloc((unsigned)(len * sizeof(struct sortel)));
	if (sortbufp == (struct sortel *)NULL)
		return;
	if (objectType != PERSON)
		return; /* only doing person at present */
	else
#ifdef USELDAP
		if (strncmp(pstr, SHORT_OU, strlen(SHORT_OU)) == 0)
#else
		if (strncmp((rindex(pstr, '@') + 1), SHORT_OU,
		    strlen(SHORT_OU)) == 0)
#endif
			dept = TRUE;
		else
			dept = FALSE;
	for (sbp = sortbufp, i = 0, x = *lp; x != NULLLIST; i++, x = x->next)
	{
		sbp->lp = x;
		sbp->key = makeKey(x, PERSON, dept);
		sbp++;
	}
	qsort((char *) sortbufp, len, sizeof(struct sortel), compareKeys);
	/* resequence list */
	reSeqList(lp, len);
	for (sbp = sortbufp, i = 0; i < len; i++, sbp++)
		free(sbp->key);
	free((char *)sortbufp);
}

